/*  IPC, Posix Shared Memory example
 *  (C) Copyright XI 2004 Michal 'GiM' Spadlinski <gim at ........ dot pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License Version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/wait.h>
#include <errno.h>

#define POSIX_SHM_NAME "/gimusia"

int main (int argc, char **argv)
{
	int             fd, pid, l_s;
	unsigned       *addr;

	l_s = sizeof(long);
	if (argc != 1) {
		shm_unlink (POSIX_SHM_NAME);
		return 0;
	}

	if ((fd = shm_open (POSIX_SHM_NAME, O_RDWR|O_CREAT, 0777)) == -1) {
		perror("[+] shm_open failed");
		return -1;
	}
	printf("[+] Posix shared mem object created\n");

	if (ftruncate (fd, l_s) == -1) {
		shm_unlink(POSIX_SHM_NAME);
		perror("[+] truncate failed");
		return -1;
	}

	addr = mmap (0, l_s, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
	if (addr == MAP_FAILED) {
		shm_unlink(POSIX_SHM_NAME);
		perror("[+] mmap failed");
		return -1;
	}
	printf("[+] Memory mmapped\n");

	shm_unlink (POSIX_SHM_NAME);
	printf ("[+] Mapped memory address is %08X\n", addr);

	fflush (stdout);

	*addr = '0';

	pid = fork ();
	if (pid == -1)
	{
		perror("[+] fork failed\n");
		shm_unlink(POSIX_SHM_NAME);
		munmap(addr, l_s);
		return -1;
	} else if (!!pid)
		fprintf(stderr, "[+] Child created\n"
				"     1 - child writes to memory\n"
				"     0 - parent writes to memory\n");
	else
		sched_yield ();

	for (l_s=0; l_s<800; l_s++)
		if (*addr == '0'+!!pid)
			putc (*addr = ('0'+!pid), stderr);
		else
			sched_yield ();

	if (!pid)
		return 0;

	waitpid(pid, NULL, 0);
	printf("[+] Child exited\n");
	munmap(addr, l_s);

	return 0;
}
